/*
 * Copyright (c) 2003-2006, 2011 Sendmail, Inc. and its suppliers.
 *	All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 *
 */

/*
**  This code contains parts of Lutz Jaenicke's TLS patch for postfix.
**  http://www.aet.tu-cottbus.de/personen/jaenicke/pfixtls/
*/

#include "sm/generic.h"
SM_RCSID("@(#)$Id: smtls.c,v 1.134 2014/03/05 00:21:18 ca Exp $")
#include "sm/assert.h"
#include "sm/error.h"
#include "sm/io.h"
#define TLSL_LOG_DEFINES 1
#include "sm/tls.h"
#include "sm/tlsmisc.h"
#include "sm/tlscnf.h"

#if MTA_USE_TLS

/*

notes:
- SSL_CTX: "global" SSL context
- SSL: session SSL context, i.e., SSL is a "child" off SSL_CTX


- use SSL extension data instead of passing tls_ctx around?
some callback functions don't have an explicit user context, e.g.,
tmp_rsa_key(), apps_ssl_info_cb().
check other programs how to do that.

*/

/*
 *  Used by applications to store data (s[cs]_sess) in SSL context.
 *  Currently only used by smtpc, not smtps.
 */
int TLSctxidx = -1;
int TLSsslidx = -1;

#define TLS_SET_STR(s, str, maxlen) do {	\
		if ((s) != NULL) {					\
			int ls;							\
			ls = strlen(s);					\
			if ((str) == NULL)				\
				(str) = sm_str_new(NULL, ls + 1, SM_MAX(ls + 2, maxlen));	\
			if ((str) != NULL)				\
				(void) sm_str_scopy((str), (s));	\
		}							\
	} while (0)

#if !TLS_NO_RSA
static RSA *rsa_tmp = NULL;	/* temporary RSA key */
static RSA *tmp_rsa_key(SSL *con, int _export, int k_eylength);
# define RSA_KEYLENGTH	512
#endif

#define bitset(bit, word)	(((bit) & (word)) != 0)

#if !NO_DH

/*
 *  Generated via: openssl dhparam -C -2 bits
 *  Maybe generate with configure and include here?
 */

static DH *get_dh512(void);

static uchar dh512_p[] = {
	0xF5,0x2A,0xFF,0x3C,0xE1,0xB1,0x29,0x40,0x18,0x11,0x8D,0x7C,
	0x84,0xA7,0x0A,0x72,0xD6,0x86,0xC4,0x03,0x19,0xC8,0x07,0x29,
	0x7A,0xCA,0x95,0x0C,0xD9,0x96,0x9F,0xAB,0xD0,0x0A,0x50,0x9B,
	0x02,0x46,0xD3,0x08,0x3D,0x66,0xA4,0x5D,0x41,0x9F,0x9C,0x7C,
	0xBD,0x89,0x4B,0x22,0x19,0x26,0xBA,0xAB,0xA2,0x5E,0xC3,0x55,
	0xE9,0x2A,0x05,0x5F
};
static uchar dh512_g[] = {
	0x02
};

static DH *
get_dh512(void)
{
	DH *dh = NULL;

	if ((dh = DH_new()) == NULL)
		return NULL;
	dh->p = BN_bin2bn(dh512_p, sizeof(dh512_p), NULL);
	dh->g = BN_bin2bn(dh512_g, sizeof(dh512_g), NULL);
	if ((dh->p == NULL) || (dh->g == NULL)) {
		DH_free(dh);
		return(NULL);
	}
	return dh;
}

static unsigned char dh1024_p[]={
	0xF4,0x88,0xFD,0x58,0x4E,0x49,0xDB,0xCD,0x20,0xB4,0x9D,0xE4,
	0x91,0x07,0x36,0x6B,0x33,0x6C,0x38,0x0D,0x45,0x1D,0x0F,0x7C,
	0x88,0xB3,0x1C,0x7C,0x5B,0x2D,0x8E,0xF6,0xF3,0xC9,0x23,0xC0,
	0x43,0xF0,0xA5,0x5B,0x18,0x8D,0x8E,0xBB,0x55,0x8C,0xB8,0x5D,
	0x38,0xD3,0x34,0xFD,0x7C,0x17,0x57,0x43,0xA3,0x1D,0x18,0x6C,
	0xDE,0x33,0x21,0x2C,0xB5,0x2A,0xFF,0x3C,0xE1,0xB1,0x29,0x40,
	0x18,0x11,0x8D,0x7C,0x84,0xA7,0x0A,0x72,0xD6,0x86,0xC4,0x03,
	0x19,0xC8,0x07,0x29,0x7A,0xCA,0x95,0x0C,0xD9,0x96,0x9F,0xAB,
	0xD0,0x0A,0x50,0x9B,0x02,0x46,0xD3,0x08,0x3D,0x66,0xA4,0x5D,
	0x41,0x9F,0x9C,0x7C,0xBD,0x89,0x4B,0x22,0x19,0x26,0xBA,0xAB,
	0xA2,0x5E,0xC3,0x55,0xE9,0x2F,0x78,0xC7,
};
static unsigned char dh1024_g[]={
	0x02,
};

static DH *
get_dh1024(void)
{
	DH *dh;

	if ((dh=DH_new()) == NULL) return(NULL);
	dh->p=BN_bin2bn(dh1024_p,sizeof(dh1024_p),NULL);
	dh->g=BN_bin2bn(dh1024_g,sizeof(dh1024_g),NULL);
	if ((dh->p == NULL) || (dh->g == NULL)) {
		DH_free(dh);
		return(NULL);
	}
	return(dh);
}

static unsigned char dh2048_p[]={
	0xCA,0x45,0x36,0x9C,0xF1,0xBA,0x79,0x15,0x88,0x25,0xDE,0xFE,
	0x7D,0x85,0x9E,0x2D,0xE6,0x4E,0xF6,0x71,0xBC,0x1F,0x9F,0x39,
	0x7E,0x57,0x9F,0x2F,0xC9,0xFA,0xD6,0xF5,0x83,0x5B,0x90,0xC6,
	0x12,0x70,0x34,0x30,0x0E,0x6B,0xCF,0x28,0x66,0x7F,0x34,0xEF,
	0x52,0x31,0x15,0x88,0x26,0xBA,0xA3,0xEF,0x39,0x7E,0x2D,0xE8,
	0x0D,0xDC,0x22,0x28,0x87,0x84,0x72,0x04,0x6D,0x37,0xBD,0x81,
	0x10,0x95,0x4B,0x5B,0x08,0x9E,0xAC,0x3C,0xD1,0x55,0x3E,0xDF,
	0x9F,0xF7,0x0D,0x9E,0xB9,0x95,0xC5,0xAF,0x3B,0x59,0xBA,0x47,
	0x40,0x49,0x8E,0xA3,0xB7,0x24,0x53,0xB6,0xF6,0xFC,0xC4,0xD0,
	0x9B,0xB8,0xB6,0x15,0xA9,0xD1,0x54,0x28,0x74,0x6D,0x6C,0x6C,
	0x6A,0x5D,0x84,0x90,0x61,0xFA,0xBD,0x85,0xDF,0x65,0xFA,0xB7,
	0x6B,0x3C,0xC5,0x38,0x60,0x3D,0xF7,0x8A,0xC2,0x8E,0xCF,0xA1,
	0x21,0xE2,0xBC,0x30,0xAE,0x73,0xF4,0x8A,0x31,0xD4,0x80,0xFD,
	0xF8,0x73,0x97,0xC6,0xEC,0x2C,0x2E,0x02,0xA6,0x09,0x70,0x11,
	0x60,0xEF,0x3E,0x4C,0x4D,0x1B,0xDF,0x9E,0x38,0x7A,0xDC,0x57,
	0xA5,0x8B,0x4E,0x99,0xE0,0x91,0x04,0x88,0xF3,0xE5,0x23,0xD8,
	0x3C,0xF9,0x2E,0xAC,0x3B,0xFE,0x3C,0xC0,0xB7,0x68,0x77,0xB0,
	0x1D,0x83,0x85,0x3B,0xE6,0x44,0x4D,0x97,0xF0,0xEE,0xDF,0xD7,
	0x21,0xA9,0x5E,0x39,0x9E,0x2C,0x9F,0xA0,0xE7,0x79,0xCA,0x14,
	0xAC,0x94,0x4D,0xD1,0x94,0xB1,0xE8,0xEF,0xA2,0xCC,0xB1,0xF5,
	0x85,0x63,0xF9,0xF7,0xE7,0x70,0x17,0x96,0x57,0xD8,0xBE,0xAF,
	0xAB,0xF0,0x2C,0xAB,
};
static unsigned char dh2048_g[]={
	0x02,
};


static DH *
get_dh2048(void)
{
	DH *dh;

	if ((dh=DH_new()) == NULL) return(NULL);
	dh->p=BN_bin2bn(dh2048_p,sizeof(dh2048_p),NULL);
	dh->g=BN_bin2bn(dh2048_g,sizeof(dh2048_g),NULL);
	if ((dh->p == NULL) || (dh->g == NULL)) {
		DH_free(dh);
		return(NULL);
	}
	return(dh);
}


#endif /* !NO_DH */

/*
**  TLSLOGERR -- log the errors from the TLS error stack
**
**	Parameters:
**		ssl -- TLS connection structure
**		tlsl_lctx -- logging context.
**		who -- server/client (for logging).
**
**	Returns:
**		none.
*/

void
tlslogerr(SSL *ssl, sm_log_ctx_P tlsl_lctx, char *who)
{
	ulong l;
	int line, flags;
	ulong es;
	const char *file, *data;
	char buf[256];
	tlsi_ctx_P tlsi_ctx;

	if (ssl != NULL)
		tlsi_ctx = (tlsi_ctx_P) SSL_get_ex_data(ssl, TLSsslidx);
	else
		tlsi_ctx = NULL;
	es = CRYPTO_thread_id();
	while ((l = ERR_get_error_line_data(&file, &line, &data, &flags))!= 0) {
		ERR_error_string_n(l, buf, sizeof(buf));
		sm_log_entry(tlsl_lctx,
			TL_LCAT_INIT, TL_LMOD_CONF, SM_LOG_WARN, 8,
			SLT_SEV, "WARN",
			SLT(SLT_FREE, 's')|SLT_COND, who != NULL, "STARTTLS", who,
			SLT(SLT_FREE, 's')|SLT_COND,
				tlsi_ctx != NULL && !SM_IS_EMPTY_S(tlsi_ctx->tlsi_log_field),
				tlsi_ctx != NULL ? tlsi_ctx->tlsi_log_field : "null-field",
				tlsi_ctx != NULL ? tlsi_ctx->tlsi_log_value : "null-value",
			SLT(SLT_FREE, 'u')|SLT_L, "thread_id", es,
			SLT(SLT_FREE, 's'), "description", buf,
			SLT(SLT_FREE, 's'), "file", file,
			SLT(SLT_FREE, 'd'), "line", line,
			SLT(SLT_FREE, 's')|SLT_COND, bitset(ERR_TXT_STRING, flags), "text", data,
			SLT_END);
	}
}

#if MTA_TLS_DEBUG
/*
**  APPS_SSL_INFO_CB -- info callback for TLS connections
**
**	Parameters:
**		ssl -- TLS connection structure
**		where -- state in handshake
**		ret -- return code of last operation
**
**	Returns:
**		none.
*/

static void
apps_ssl_info_cb(const SSL *ssl, int where, int ret)
{
	int w;
	char *str;
	tlsi_ctx_P tlsi_ctx;
	sm_log_ctx_P tlsl_lctx;

	tlsl_lctx = NULL;
	tlsi_ctx = NULL;
	if (ssl != NULL) {
		tlsi_ctx = (tlsi_ctx_P) SSL_get_ex_data(ssl, TLSsslidx);
		if (tlsi_ctx != NULL) {
			tls_ctx_P tls_ctx;

			tls_ctx = tlsi_ctx->tlsi_tls;
			if (tls_ctx != NULL)
				tlsl_lctx = tls_ctx->tls_lctx;
		}
	}

	w = where & ~SSL_ST_MASK;
	if (bitset(SSL_ST_CONNECT, w))
		str = "SSL_connect";
	else if (bitset(SSL_ST_ACCEPT, w))
		str = "SSL_accept";
	else
		str = "undefined";

	if (bitset(SSL_CB_LOOP, where)) {
		sm_log_write(tlsl_lctx,
			TL_LCAT_IO, TL_LMOD_IO, SM_LOG_INFO, 11,
			"func=apps_ssl_info_cb, %s=%@s", str, SSL_state_string_long(ssl));
	}
	else if (bitset(SSL_CB_ALERT, where)) {
		str = bitset(SSL_CB_READ, where) ? "read" : "write";
		sm_log_write(tlsl_lctx,
			TL_LCAT_IO, TL_LMOD_IO, SM_LOG_INFO, 11,
			"func=apps_ssl_info_cb, SSL3_alert=%s, alert_type=%@s, alert_descr=%@s",
			str, SSL_alert_type_string_long(ret), SSL_alert_desc_string_long(ret));
	}
	else if (bitset(SSL_CB_EXIT, where)) {
		if (ret == 0) {
			sm_log_write(tlsl_lctx,
				TL_LCAT_IO, TL_LMOD_IO, SM_LOG_INFO, 11,
				"func=apps_ssl_info_cb, %s=failed, state=%@s",
				str, SSL_state_string_long(ssl));
		}
		else if (ret < 0) {
#if !MTA_TLS_DEBUG_ALL
			switch (SSL_get_error(ssl, ret)) {
			  case SSL_ERROR_WANT_READ:
			  case SSL_ERROR_WANT_WRITE:
				break;
			  default:
#endif
				sm_log_write(tlsl_lctx,
					TL_LCAT_IO, TL_LMOD_IO, SM_LOG_ERROR, 9,
					"func=apps_ssl_info_cb, %s=error, state=%@s",
					str, SSL_state_string_long(ssl));
#if !MTA_TLS_DEBUG_ALL
			}
#endif
		}
	}
}
#endif /* MTA_TLS_DEBUG */

/*
**  TLS_VERIFY_CERT_CB -- verify callback for TLS certificates
**
**	Parameters:
**		ok -- ok so far? (0: not ok)
**		ctx -- x509 context
**
**	Returns:
**		accept connection?
**		currently: always yes.
**
**	Note:
**		This might be called several times for the same "depth"
**		(e.g., twice for top-level CA?once with ok=0, then again with ok=1?
**		because of the override?)
*/

static int
tls_verify_cert_cb(int ok, X509_STORE_CTX *ctx)
{
	SSL *ssl;
	X509 *cert;
	int reason, depth, max_vrfy_depth;
	bool dok;
	tls_ctx_P tls_ctx;
	tlsi_ctx_P tlsi_ctx;
	const char *errtxt;
	char buf[SM_X509_NAME_MAX];

	dok = false;
	cert = X509_STORE_CTX_get_current_cert(ctx);
	reason = X509_STORE_CTX_get_error(ctx);
	depth = X509_STORE_CTX_get_error_depth(ctx);
	ssl = (SSL *) X509_STORE_CTX_get_ex_data(ctx,
			SSL_get_ex_data_X509_STORE_CTX_idx());
	if (NULL == ssl) {
		sm_log_write(NULL,	/* NO log context !*/
			TL_LCAT_INIT, TL_LMOD_CONF, SM_LOG_ERROR, 9,
			"STARTTLS=internal_error, func=tls_verify_log, ssl=NULL");
		return 0;	/* really abort handshake? */
	}

	tlsi_ctx = (tlsi_ctx_P) SSL_get_ex_data(ssl, TLSsslidx);
	if (NULL == tlsi_ctx) {
		sm_log_write(NULL,	/* NO log context !*/
			TL_LCAT_INIT, TL_LMOD_CONF, SM_LOG_ERROR, 9,
			"STARTTLS=internal_error, func=tls_verify_log, tlsi_ctx=NULL");
		return 0;	/* really abort handshake? */
	}
	tls_ctx = tlsi_ctx->tlsi_tls;
	/* SM_ASSERT(NULL != tls_ctx); */

	if (cert != NULL) {
		X509_NAME_oneline(X509_get_subject_name(cert), buf, sizeof buf);
		dok = depth >= 0 && depth < SM_ARRAY_SIZE(tlsi_ctx->tlsi_ca_subjects);
		if (dok)
			TLS_SET_XSTR(buf, tlsi_ctx->tlsi_ca_subjects[depth]);
	}
	else {
		sm_log_entry(tls_ctx->tls_lctx,
			TL_LCAT_INIT, TL_LMOD_CONF, SM_LOG_INFO, ok ? 14 : 9,
			SLT_SEV, "ERROR",
			SLT_FUNC, "tls_verify_cert_cb",
			SLT(SLT_FREE, 's')|SLT_COND, !SM_IS_EMPTY_S(tlsi_ctx->tlsi_log_field), tlsi_ctx->tlsi_log_field, tlsi_ctx->tlsi_log_value,
			SLT(SLT_FREE, 's'), "cert", "NULL");
	}

	/*
	 * Catch a too long certificate chain. The depth limit set using
	 * SSL_CTX_set_verify_depth() is by purpose set to "limit+1" so
	 * that whenever the "depth + 1>verify_depth" condition is met, we
	 * have violated the limit and want to log this error condition.
	 * We must do it here, because the CHAIN_TOO_LONG error would not
	 * be found explicitly; only errors introduced by cutting off the
	 * additional certificates would be logged.
	 * (based on SSL_CTX_set_verify(3))
	 */

	max_vrfy_depth = SSL_get_verify_depth(ssl);
	if (max_vrfy_depth >= 0 && depth + 1 > max_vrfy_depth) {
		reason = X509_V_ERR_CERT_CHAIN_TOO_LONG;
		X509_STORE_CTX_set_error(ctx, reason);
	}
#define EVP_DIGEST	tls_ctx->tls_EVP_digest

	errtxt = X509_verify_cert_error_string(reason);
	if (reason != X509_V_OK && SM_IS_EMPTY_STR(tlsi_ctx->tlsi_vrfy_err))
		TLS_SET_STR(errtxt, tlsi_ctx->tlsi_vrfy_err, SM_X509_NAME_MAX);
	if (tls_ctx != NULL) {
		uchar md[EVP_MAX_MD_SIZE];
		uint n;

		md[0] = '\0';
		n = 0;
		if (sm_log_wouldlog(tls_ctx->tls_lctx, TL_LCAT_INIT, TL_LMOD_CONF, ok ? 15 : 10) && cert != NULL)
		{
			if (X509_digest(cert, EVP_DIGEST, md, &n) == 0)
				n = 0;
		}

		sm_log_entry(tls_ctx->tls_lctx,
			TL_LCAT_INIT, TL_LMOD_CONF, SM_LOG_INFO, ok ? 14 : 9,
			SLT_SEV, "INFO",
			SLT_FUNC, "tls_verify_cert_cb",
			SLT(SLT_FREE, 's')|SLT_COND, !SM_IS_EMPTY_S(tlsi_ctx->tlsi_log_field), tlsi_ctx->tlsi_log_field, tlsi_ctx->tlsi_log_value,
			SLT(SLT_FREE, 'd'), "depth", depth,
			SLT_CERT_SUB|SLT_COND, dok, dok ? tlsi_ctx->tlsi_ca_subjects[depth] : NULL,
			SLT(SLT_FREE, 's')|SLT_XTEXT|SLT_COND, cert != NULL, "cert_issuer", X509_NAME_oneline(X509_get_issuer_name(cert), buf, sizeof buf),
			SLT(SLT_FREE, 'y')|SLT_PREC|SLT_C|SLT_COND|SLT_ZEROPAD, n > 0, (int) n, "cert_fp", md,
			SLT_STAT|SLT_COND, ok != 0, ok,
			SLT_STATUS|SLT_CLEAN, errtxt,
			SLT_END);
	}
	return 1;
}

/*
**  X509_VERIFY_CB -- verify callback for CRL
**
**	Parameters:
**		ok -- ok?
**		ctx -- x509 context
**
**	Returns:
**		accept CRL content?
**		currently: always yes.
*/

static int
x509_verify_cb(int ok, X509_STORE_CTX *ctx)
{
	if (0 == ok) {
		if (X509_V_ERR_UNABLE_TO_GET_CRL == ctx->error) {
			ctx->error = 0;
			return 1;
		}
	}
	return ok;
}


/*
**  SM_TLS_INIT_LIBRARY -- setup TLS library for global use
**
**	Parameters:
**		ptls_ctx -- (pointer to) TLS library context (output)
**		fipsmode -- FIPS mode
**
**	Returns:
**		succeeded?
*/

sm_ret_T
sm_tls_init_library(tls_ctx_P *ptls_ctx, bool fipsmode)
{
	sm_ret_T ret;
	tls_ctx_P tls_ctx;

	SM_REQUIRE(ptls_ctx != NULL);
	*ptls_ctx = NULL;
	tls_ctx = (tls_ctx_P) sm_zalloc(sizeof(*tls_ctx));
	if (NULL == tls_ctx)
		return sm_error_temp(SM_EM_TLS, ENOMEM);

	/* basic TLS initialization, ignore result for now */
	SSL_library_init();
	SSL_load_error_strings();
	OpenSSL_add_all_algorithms();
	if (TLSctxidx < 0) {
		TLSctxidx = SSL_get_ex_new_index(0, 0, 0, 0, 0);
		if (TLSctxidx < 0) {
			sm_log_write(tls_ctx->tls_lctx,
				TL_LCAT_INIT, TL_LMOD_CONF, SM_LOG_ERROR, 2,
				"STARTTLS=init, SSL_get_ex_new_index=%d", TLSctxidx);
			return sm_error_temp(SM_EM_TLS, EINVAL);
		}
	}
	if (TLSsslidx < 0) {
		TLSsslidx = SSL_get_ex_new_index(0, 0, 0, 0, 0);
		if (TLSsslidx < 0) {
			sm_log_write(tls_ctx->tls_lctx,
				TL_LCAT_INIT, TL_LMOD_CONF, SM_LOG_ERROR, 2,
				"STARTTLS=init, SSL_get_ex_new_index=%d", TLSsslidx);
			return sm_error_temp(SM_EM_TLS, EINVAL);
		}
	}

	/* this should be done before the logging above... */
	ret = sm_log_create(NULL, &tls_ctx->tls_lctx, &tls_ctx->tls_lcfg);
	if (sm_is_err(ret))
		goto error;
	ret = sm_log_setfile(tls_ctx->tls_lctx, smioerr);
	if (sm_is_err(ret))
		goto error;

	/* log level can be changed in sm_set_tls_log() */
	ret = sm_log_setdebuglevel(tls_ctx->tls_lctx, 9);
	if (sm_is_err(ret))
		goto error;

	tls_ctx->tls_EVP_digest = EVP_sha1();

# if MTA_USE_TLSFIPS
	if (fipsmode) {
		if (!FIPS_mode_set(1)) {
			unsigned long err;

			err = ERR_get_error();
			sm_log_write(tls_ctx->tls_lctx,
				TL_LCAT_INIT, TL_LMOD_CONF, SM_LOG_ERROR, 2,
				"STARTTLS=init, FIPS_mode=%s", ERR_error_string(err, NULL));
			return sm_error_temp(SM_EM_TLS, EINVAL);
		}
		else
			sm_log_write(tls_ctx->tls_lctx,
				TL_LCAT_INIT, TL_LMOD_CONF, SM_LOG_INFO, 8,
				"STARTTLS=init, FIPS_mode=ok");
	}
# endif

	*ptls_ctx = tls_ctx;
	return SM_SUCCESS;

  error:
	/* cleanup? */
	return ret;
}

/*
**  SM_SET_TLS_LOG -- set TLS logging
**
**	Parameters:
**		tls_ctx -- TLS library context
**		fp -- file to use
**		fd -- fd to use
**		loglevel -- loglevel
**
**	Returns:
**		succeeded?
*/

sm_ret_T
sm_set_tls_log(tls_ctx_P tls_ctx, sm_file_T *fp, int fd, int loglevel)
{
	sm_ret_T ret;

	ret = sm_log_setfp_fd(tls_ctx->tls_lctx, fp, fd);
	if (sm_is_err(ret))
		goto error;
	ret = sm_log_setdebuglevel(tls_ctx->tls_lctx, loglevel);
	if (sm_is_err(ret))
		goto error;
	return SM_SUCCESS;

  error:
	return ret;
}

/*
**  status in initialization
**  these flags keep track of the status of the initialization
**  i.e., whether a file exists (_EX) and whether it can be used (_OK)
**  [due to permissions]
*/

#define TLS_S_NONE		0x00000000	/* none yet */
#define TLS_S_CERT_OK(i)	(0x00000001 << (i << 1))	/* cert file is ok */
#define TLS_S_KEY_OK(i)	(0x00000002 << (i << 1))	/* key file is ok */

#define TLS_S_CERTP_OK	0x00010000	/* CA cert path is ok */
#define TLS_S_CERTF_OK	0x00040000	/* CA cert file is ok */
#define TLS_S_LISTF_OK	0x00080000	/* CA list file is ok */

#define TLS_S_DHPAR_OK	0x00800000	/* DH param file is ok to use */

/*
**  TLS_SET_VERIFY -- request certificate?
**
**	Parameters:
**		ctx -- TLS context
**		ssl -- TLS connection structure
**		req -- request certificate?
**
**	Returns:
**		none.
**
**	Side Effects:
**		Sets verification state for TLS connection
**
*/

void
tls_set_verify(SSL_CTX *ctx, SSL *ssl, bool req)
{
	SSL_set_verify(ssl, req ? SSL_VERIFY_PEER : SSL_VERIFY_NONE, tls_verify_cert_cb);
}

/*
**  TLS_SET_EECDH_CURVE -- set EC curve
**
**	Parameters:
**		tls_ctx -- TLS library context
**		ctx -- SSL context
**		curve -- EC curve
**
**	Returns:
**		usual error code
*/

static sm_ret_T
tls_set_eecdh_curve(tls_ctx_P tls_ctx, SSL_CTX *ctx, const char *curve)
{
# if MTA_HAVE_ECDSA
	int r, nid;
	EC_KEY *ecdh;

	/* sm_io_fprintf(smioerr, "curve=%@s\n", curve); */
	if (SM_IS_EMPTY_S(curve))
		return SM_NOTDONE;

	/*
	 * Elliptic-Curve Diffie-Hellman parameters are either "named curves"
	 * from RFC 4492 section 5.1.1, or explicitly described curves over
	 * binary fields. OpenSSL only supports the "named curves", which provide
	 * maximum interoperability. The recommended curve for 128-bit
	 * work-factor key exchange is "prime256v1" a.k.a. "secp256r1" from
	 * Section 2.7 of http://www.secg.org/download/aid-386/sec2_final.pdf
	 */

	if ((nid = OBJ_sn2nid(curve)) == NID_undef)
		return sm_err_perm(EINVAL);
	ERR_clear_error();
	ecdh = EC_KEY_new_by_curve_name(nid);
	if (NULL == ecdh) {
		sm_log_write(tls_ctx->tls_lctx,
			TL_LCAT_INIT, TL_LMOD_CONF, SM_LOG_WARN, 7,
			"STARTTLS=server, EC_KEY_new_by_curve_name=failed, curve=%@s",
			curve);
		tlslogerr(NULL, tls_ctx->tls_lctx, "srv");
		return sm_err_perm(EINVAL);
	}

	SSL_CTX_set_options(tls_ctx->tls_ssl_ctx, SSL_OP_SINGLE_ECDH_USE);
	if ((r = SSL_CTX_set_tmp_ecdh(ctx, ecdh)) == 0) {
		sm_log_write(tls_ctx->tls_lctx,
			TL_LCAT_INIT, TL_LMOD_CONF, SM_LOG_WARN, 7,
			"STARTTLS=server, SSL_CTX_set_tmp_ecdh=failed, curve=%@s, stat=%r",
			curve, r);
		tlslogerr(NULL, tls_ctx->tls_lctx, "srv");
		return sm_err_perm(EINVAL);
	}
	return SM_SUCCESS;
#else
	return SM_NOTDONE;
#endif /* MTA_HAVE_ECDSA */
}

/*
**  SM_LOADCERT -- load cert and key
**
**	Parameters:
**		tls_ctx -- TLS library context
**		ctx -- context
**		req -- requirements for initialization
**		status -- current result of status checks
**		srv -- server side?
**		idx -- index
**		keyfile -- keyfile
**		certfile -- certfile
**
**	Returns:
**		usual error code
*/

static sm_ret_T
sm_loadcert(tls_ctx_P tls_ctx, SSL_CTX *ctx, ulong req, long status, bool srv, uint idx, const char *keyfile, const char *certfile)
{
	int r;
	char *who;

	who = srv ? "server" : "client";

	/* usable cert and key? */
	if (!(bitset(TLS_S_CERT_OK(idx), status) &&
		 (bitset(TLS_S_KEY_OK(idx), status)))) {
		return SM_SUCCESS;
	}

	/* get the certificate file */
	r = 1;
	if ((r = SSL_CTX_use_certificate_chain_file(ctx, certfile)) <= 0) {
		sm_log_write(tls_ctx->tls_lctx,
			TL_LCAT_INIT, TL_LMOD_CONF, SM_LOG_WARN, 7,
			"STARTTLS=%s, SSL_CTX_use_certificate_file=failed, certfile=%s, ret=%d",
			who, certfile, r);
		tlslogerr(NULL, tls_ctx->tls_lctx, who);
		if (bitset(TLS_I_USE_CERT, req))
			return sm_error_perm(SM_EM_TLS, EINVAL);
	}
	/* sm_io_fprintf(smioerr, "STARTTLS=%s, SSL_CTX_use_certificate_chain_file=%d, certfile=%s\n", who, r, certfile); */

	/* load private key */
	if ((r = SSL_CTX_use_PrivateKey_file(ctx, keyfile, SSL_FILETYPE_PEM)) <= 0)
	{
		sm_log_write(tls_ctx->tls_lctx,
			TL_LCAT_INIT, TL_LMOD_CONF, SM_LOG_WARN, 7,
			"STARTTLS=%s, SSL_CTX_use_PrivateKey_file=failed, certfile=%s, keyfile=%s, ret=%d",
			who, certfile, keyfile, r);
		tlslogerr(NULL, tls_ctx->tls_lctx, who);
		if (bitset(TLS_I_USE_KEY, req))
			return sm_error_perm(SM_EM_TLS, EINVAL);
	}
	/* sm_io_fprintf(smioerr, "STARTTLS=%s, SSL_CTX_use_PrivateKey_file=%d, certfile=%s, keyfile=%s\n", who, r, certfile, keyfile); */

	/* check the private key */
	if ((r = SSL_CTX_check_private_key(ctx)) <= 0) {
		/* Private key does not match the certificate public key */
		sm_log_write(tls_ctx->tls_lctx,
			TL_LCAT_INIT, TL_LMOD_CONF, SM_LOG_WARN, 7,
			"STARTTLS=%s, SSL_CTX_check_private_key=failed, certfile=%s, keyfile=%s, error=%d",
			who, certfile, keyfile, r);
		tlslogerr(NULL, tls_ctx->tls_lctx, who);
		if (bitset(TLS_I_USE_KEY, req))
			return sm_error_perm(SM_EM_TLS, EINVAL);
	}
	/* sm_io_fprintf(smioerr, "STARTTLS=%s, SSL_CTX_check_private_key=%d, certfile=%s, keyfile=%s\n", who, r, certfile, keyfile); */
	return SM_SUCCESS;
}

/*
**  SM_SSL_LOADCERT -- load cert and key for a TLS session
**
**	Parameters:
**		tls_ctx -- TLS library context
**		SSL -- TLS session context
**		srv -- server side?
**		keyfile -- keyfile
**		certfile -- certfile
**
**	Returns:
**		usual error code
*/

sm_ret_T
sm_ssl_loadcert(tls_ctx_P tls_ctx, SSL *ssl, bool srv, const char *keyfile, const char *certfile)
{
	int r;
	char *who;

	SM_REQUIRE(ssl != NULL);
	SM_REQUIRE(keyfile != NULL);
	SM_REQUIRE(certfile != NULL);

	who = srv ? "server" : "client";

	if ((r = SSL_use_certificate_file(ssl, certfile, SSL_FILETYPE_PEM)) <= 0) {
		sm_log_write(tls_ctx->tls_lctx,
			TL_LCAT_INIT, TL_LMOD_CONF, SM_LOG_WARN, 7,
			"STARTTLS=%s, SSL_use_certificate_file=failed, certfile=%s, ret=%d",
			who, certfile, r);
		tlslogerr(ssl, tls_ctx->tls_lctx, who);
		return sm_error_perm(SM_EM_TLS, EINVAL);
	}
	if ((r = SSL_use_PrivateKey_file(ssl, keyfile, SSL_FILETYPE_PEM)) <= 0)
	{
		sm_log_write(tls_ctx->tls_lctx,
			TL_LCAT_INIT, TL_LMOD_CONF, SM_LOG_WARN, 7,
			"STARTTLS=%s, SSL_use_PrivateKey_file=failed, certfile=%s, keyfile=%s, ret=%d",
			who, certfile, keyfile, r);
		tlslogerr(ssl, tls_ctx->tls_lctx, who);
		return sm_error_perm(SM_EM_TLS, EINVAL);
	}
	/* sm_io_fprintf(smioerr, "STARTTLS=%s, SSL_use_PrivateKey_file=%d, certfile=%s, keyfile=%s\n", who, r, certfile, keyfile); */

	/* check the private key */
	if ((r = SSL_check_private_key(ssl)) <= 0) {
		/* Private key does not match the certificate public key */
		sm_log_write(tls_ctx->tls_lctx,
			TL_LCAT_INIT, TL_LMOD_CONF, SM_LOG_WARN, 7,
			"STARTTLS=%s, SSL_check_private_key=failed, certfile=%s, keyfile=%s, error=%d",
			who, certfile, keyfile, r);
		tlslogerr(ssl, tls_ctx->tls_lctx, who);
	}
	/* sm_io_fprintf(smioerr, "STARTTLS=%s, SSL_check_private_key=%d, certfile=%s, keyfile=%s\n", who, r, certfile, keyfile); */
	return SM_SUCCESS;
}

/*
**  SM_TLS_INIT -- initialize TLS
**
**	Parameters:
**		tls_ctx -- TLS library context
**		ctx -- pointer to context (output)
**		req -- requirements for initialization
**		srv -- server side?
**		tls_cnf -- TLS configuration data
**
**	Returns:
**		succeeded?
*/

/*
**  The session_id_context identifies the service that created a session.
**  This information is used to distinguish between multiple TLS-based
**  servers running on the same server. We use the name of the mail system.
**  Note: the session cache is not persistent.
*/

static char server_session_id_context[] = "MeTA1";

/* 0.9.8a and b have a problem with SSL_OP_TLS_BLOCK_PADDING_BUG */
#if (OPENSSL_VERSION_NUMBER >= 0x0090800fL)
# define SM_SSL_OP_TLS_BLOCK_PADDING_BUG	1
#else
# define SM_SSL_OP_TLS_BLOCK_PADDING_BUG	0
#endif

sm_ret_T
sm_tls_init(tls_ctx_P tls_ctx, ulong req, bool srv, tls_cnf_P tls_cnf)
{
#if !NO_DH
	static DH *dh = NULL;
#endif
	int r;
	sm_ret_T ret;
	bool ok;
	long status, options;
	char *who;
	const char *param;
#if SM_SSL_OP_TLS_BLOCK_PADDING_BUG
	long rt_version;
	STACK_OF(SSL_COMP) *comp_methods;
#endif
	BIO *crl_file;
	X509_STORE *store;

	SM_REQUIRE(tls_cnf != NULL);
	SM_REQUIRE(TLSctxidx >= 0);
	status = TLS_S_NONE;
	who = srv ? "server" : "client";

	/* already initialized? (we could re-init...) */
	if (tls_ctx->tls_ssl_ctx != NULL)
		return SM_SUCCESS;
	ok = true;
	tls_ctx->tls_cnf = tls_cnf;

#define TLS_OK_F(var, fn, req, st, srv) if (ok) {	\
		r = (!SM_IS_EMPTY_S(var));	\
		if (r) \
			status |= st; \
		else if (req) \
			ok = false; \
	}

#if !NO_DH
	/*
	**  none	no parameters: don't use DH
	**  512		generate 512 bit parameters (fixed)
	**  1024	generate 1024 bit parameters
	**  /file/name	read parameters from /file/name
	**  default is: 1024 for server, 512 for client (OK? XXX)
	*/

	param = NULL;
	if (bitset(TLS_I_TRY_DH, req)) {
		dhparam_T dhparam;

		dhparam = tls_cnf->tlscnf_dhparam;
		switch (dhparam.dhparam_type) {
		  case DHPBITS:
			if (2048 == dhparam.dhbits.dhbits_bits)
				req |= TLS_I_DH2048;
			else if (1024 == dhparam.dhbits.dhbits_bits)
				req |= TLS_I_DH1024;
			else if (512 == dhparam.dhbits.dhbits_bits)
				req |= TLS_I_DH512;
			else {
				sm_log_write(tls_ctx->tls_lctx,
					TL_LCAT_INIT, TL_LMOD_CONF, SM_LOG_WARN, 12,
					"STARTTLS=%s, DHParam=%d, error=illegal_value",
					who, dhparam.dhbits.dhbits_bits);
			}
			break;
		  case DHPFILE:
			param = dhparam.dhpfile.dhpfile_name;
			TLS_OK_F(param, "DHParameters",
				 bitset(TLS_I_DHPAR_EX, req), TLS_S_DHPAR_OK, srv);
			break;
		  default:
			/* should we use a default? */
			req |= srv ? TLS_I_DH1024 : TLS_I_DH512;
			break;
		}
	}
	if (!ok)
		return sm_error_perm(SM_EM_TLS, EINVAL);
#endif /* !NO_DH */

	TLS_OK_F(tls_cnf->tlscnf_cacertpath, "CAcert_directory",
		bitset(TLS_I_CERTP_EX, req), TLS_S_CERTP_OK, srv);
	TLS_OK_F(tls_cnf->tlscnf_cacertfile, "CAcert_file",
		bitset(TLS_I_CERTF_EX, req), TLS_S_CERTF_OK, srv);
	TLS_OK_F(tls_cnf->tlscnf_calistfile, "CAlist_file",
		bitset(TLS_I_LISTF_EX, req), TLS_S_LISTF_OK, srv);

	TLS_OK_F(tls_cnf->tlscnf_certfile, "CertFile",
		bitset(TLS_I_CERT_EX, req), TLS_S_CERT_OK(TLS_IDX_RSA), srv);
	TLS_OK_F(tls_cnf->tlscnf_keyfile, "KeyFile",
		bitset(TLS_I_KEY_EX, req), TLS_S_KEY_OK(TLS_IDX_RSA), srv);

	if (tls_cnf->tlscnf_dsa_certfile != NULL &&
	    tls_cnf->tlscnf_dsa_keyfile != NULL)
	{
		TLS_OK_F(tls_cnf->tlscnf_dsa_certfile, "DSACertFile",
			bitset(TLS_I_CERT_EX, req),
			TLS_S_CERT_OK(TLS_IDX_DSA), srv ? TLS_T_SRV : TLS_T_CLT);
		TLS_OK_F(tls_cnf->tlscnf_dsa_keyfile, "DSAKeyFile",
			bitset(TLS_I_KEY_EX, req),
			TLS_S_KEY_OK(TLS_IDX_DSA), srv ? TLS_T_SRV : TLS_T_CLT);
	}

# if MTA_HAVE_ECDSA
	if (tls_cnf->tlscnf_ec_certfile != NULL &&
	    tls_cnf->tlscnf_ec_keyfile != NULL)
	{
		TLS_OK_F(tls_cnf->tlscnf_ec_certfile, "ECCertFile",
			bitset(TLS_I_CERT_EX, req),
			TLS_S_CERT_OK(TLS_IDX_EC), srv ? TLS_T_SRV : TLS_T_CLT);
		TLS_OK_F(tls_cnf->tlscnf_ec_keyfile, "ECKeyFile",
			bitset(TLS_I_KEY_EX, req),
			TLS_S_KEY_OK(TLS_IDX_EC), srv ? TLS_T_SRV : TLS_T_CLT);
	}
#endif

	/* create a method and a new context */
	tls_ctx->tls_ssl_ctx = SSL_CTX_new(srv ? SSLv23_server_method() :
				      SSLv23_client_method());
	if (NULL == tls_ctx->tls_ssl_ctx) {
		sm_log_write(tls_ctx->tls_lctx,
			TL_LCAT_INIT, TL_LMOD_CONF, SM_LOG_ERR, 7,
			"STARTTLS=%s, SSL_CTX_new(SSLv23_%s_method())=failed",
			who, who);
		tlslogerr(NULL, tls_ctx->tls_lctx, who);
		return sm_error_perm(SM_EM_TLS, EINVAL);
	}

	if (srv)
		ret = tls_set_eecdh_curve(tls_ctx, tls_ctx->tls_ssl_ctx,
				tls_cnf->tlscnf_ec_curve);

#if TLS_NO_RSA
	/* turn off backward compatibility, required for no-rsa */
	SSL_CTX_set_options(tls_ctx->tls_ssl_ctx, SSL_OP_NO_SSLv2);
#endif

#if !TLS_NO_RSA
	/*
	**  Create a temporary RSA key
	**  XXX  Maybe we shouldn't create this always (even though it
	**  is only at startup).
	**  It is a time-consuming operation and it is not always necessary.
	**  maybe we should do it only on demand...
	*/

	if (bitset(TLS_I_RSA_TMP, req) &&
	    (rsa_tmp = RSA_generate_key(RSA_KEYLENGTH, RSA_F4, NULL, NULL)) == NULL)
	{
		sm_log_write(tls_ctx->tls_lctx,
			TL_LCAT_INIT, TL_LMOD_CONF, SM_LOG_WARN, 7,
			"STARTTLS=%s, RSA_generate_key=failed", who);
		tlslogerr(NULL, tls_ctx->tls_lctx, who);
		return sm_error_perm(SM_EM_TLS, EINVAL);
	}
#endif /* !TLS_NO_RSA */

	/* load private key */
	ret = sm_loadcert(tls_ctx, tls_ctx->tls_ssl_ctx, req, status, srv, 0,
			tls_cnf->tlscnf_keyfile, tls_cnf->tlscnf_certfile);
	if (sm_is_err(ret))
		return ret;

	ret = sm_loadcert(tls_ctx, tls_ctx->tls_ssl_ctx, req, status, srv, 1,
			tls_cnf->tlscnf_dsa_keyfile, tls_cnf->tlscnf_dsa_certfile);
	if (sm_is_err(ret))
		return ret;

# if MTA_HAVE_ECDSA
	ret = sm_loadcert(tls_ctx, tls_ctx->tls_ssl_ctx, req, status, srv, 2,
			tls_cnf->tlscnf_ec_keyfile, tls_cnf->tlscnf_ec_certfile);
	if (sm_is_err(ret))
		return ret;
#endif

	/* standard violation?
	*SSL_CTX_set_quiet_shutdown(tls_ctx->tls_ssl_ctx, 1);
	*/

	if (tls_cnf->tlscnf_options_isset != 0)
		options = tls_cnf->tlscnf_options;
	else {
		options = SSL_OP_ALL;
#ifdef SSL_OP_NO_TICKET
		if (!srv && 0 == tls_cnf->tlscnf_cache_size)
			options |= SSL_OP_NO_TICKET;
#endif
	}

#if SM_SSL_OP_TLS_BLOCK_PADDING_BUG

	/*
	**  In OpenSSL 0.9.8[ab], enabling zlib compression breaks the
	**  padding bug work-around, leading to false positives and
	**  failed connections. We may not interoperate with systems
	**  with the bug, but this is better than breaking on all 0.9.8[ab]
	**  systems that have zlib support enabled.
	**  Note: this checks the runtime version of the library, not
	**  just the compile time version.
	*/

	rt_version = SSLeay();
	if (rt_version >= 0x00908000L && rt_version <= 0x0090802fL) {
		comp_methods = SSL_COMP_get_compression_methods();
		if (comp_methods != NULL && sk_SSL_COMP_num(comp_methods) > 0)
			options &= ~SSL_OP_TLS_BLOCK_PADDING_BUG;
	}
#endif
	SSL_CTX_set_options(tls_ctx->tls_ssl_ctx, options);

#if !NO_DH
	/* Diffie-Hellman initialization */
	if (bitset(TLS_I_TRY_DH, req)) {
		if (bitset(TLS_S_DHPAR_OK, status)) {
			BIO *bio;

			if ((bio = BIO_new_file(param, "r")) != NULL) {
				dh = PEM_read_bio_DHparams(bio, NULL, NULL, NULL);
				BIO_free(bio);
				if (NULL == dh) {
					ulong err;

					err = ERR_get_error();
					sm_log_write(tls_ctx->tls_lctx,
						TL_LCAT_INIT, TL_LMOD_CONF, SM_LOG_WARN, 7,
						"STARTTLS=%s, DH_parameters=%s, error=cannot read, status=%s",
						who, param, ERR_error_string(err, NULL));
					tlslogerr(NULL, tls_ctx->tls_lctx, who);
				}
			}
			else {
				sm_log_write(tls_ctx->tls_lctx,
					TL_LCAT_INIT, TL_LMOD_CONF, SM_LOG_WARN, 5,
					"STARTTLS=%s, DH_parameters=%s, BIO_new_file=failed",
					who, param);
				tlslogerr(NULL, tls_ctx->tls_lctx, who);
			}
		}
		if (dh == NULL && bitset(TLS_I_DH2048, req))
			dh = get_dh2048();
		if (dh == NULL && bitset(TLS_I_DH1024, req))
			dh = get_dh1024();
		if (dh == NULL && bitset(TLS_I_DH512, req))
			dh = get_dh512();

		if (NULL == dh) {
			ulong err;

			err = ERR_get_error();
			if (bitset(TLS_I_REQ_DH, req))
				sm_log_write(tls_ctx->tls_lctx,
					TL_LCAT_INIT, TL_LMOD_CONF, SM_LOG_WARN, 9,
					"STARTTLS=%s, DH_parameters=%s, error=cannot_read_or_set, status=%s",
					who, param, ERR_error_string(err, NULL));
			if (bitset(TLS_I_REQ_DH, req))
				return sm_error_perm(SM_EM_TLS, EINVAL);
		}
		else {
			/* important to avoid small subgroup attacks */
			SSL_CTX_set_options(tls_ctx->tls_ssl_ctx, SSL_OP_SINGLE_DH_USE);
			SSL_CTX_set_tmp_dh(tls_ctx->tls_ssl_ctx, dh);
			sm_log_write(tls_ctx->tls_lctx,
				TL_LCAT_INIT, TL_LMOD_CONF, SM_LOG_INFO, 13,
				"STARTTLS=%s, Diffie-Hellman=init, key=%d, dhparam=%s",
				who, 8 * DH_size(dh), param);
			DH_free(dh);
		}
	}
#endif /* !NO_DH */

	if (tls_cnf->tlscnf_cache_size > 0) {
		r = SSL_CTX_sess_set_cache_size(tls_ctx->tls_ssl_ctx,
				tls_cnf->tlscnf_cache_size);
		/* r == previously valid size, no error possible? */
		if (tls_cnf->tlscnf_cache_tmout > 0)
			r = SSL_CTX_set_timeout(tls_ctx->tls_ssl_ctx,
					tls_cnf->tlscnf_cache_tmout);
		/* r == previous timeout value, no error possible */

		/* server side only? */
		r = SSL_CTX_set_session_id_context(tls_ctx->tls_ssl_ctx,
			(const unsigned char *) &server_session_id_context,
			sizeof(server_session_id_context));
		if (r != 1) {
			sm_log_write(tls_ctx->tls_lctx,
				TL_LCAT_INIT, TL_LMOD_CONF, SM_LOG_ERR, 6,
				"STARTTLS=%s, SSL_CTX_set_session_id_context=failed" , who);
		}
		else if (srv)
			(void) SSL_CTX_set_session_cache_mode(tls_ctx->tls_ssl_ctx,
					SSL_SESS_CACHE_SERVER);
		else {
#ifndef SSL_SESS_CACHE_NO_INTERNAL_STORE
#define SSL_SESS_CACHE_NO_INTERNAL_STORE 0
#endif
			(void) SSL_CTX_set_session_cache_mode(tls_ctx->tls_ssl_ctx,
				 SSL_SESS_CACHE_CLIENT
				|SSL_SESS_CACHE_NO_INTERNAL_STORE
				|SSL_SESS_CACHE_NO_AUTO_CLEAR);
			tls_ctx->tls_clt_cache = true;
		}
	}
	else
		r = 0;
	if (r != 1) {
		(void) SSL_CTX_set_session_cache_mode(tls_ctx->tls_ssl_ctx,
				SSL_SESS_CACHE_OFF);
		/*
		 *  Why is this done? There is some problem with SSL_SESS_CACHE_OFF
		 *  but it's not clear what... this avoids an error
		 *  4476:error:140D9115:SSL routines:SSL_GET_PREV_SESSION:session id context uninitialized:ssl_sess.c:413
		 */
		(void) SSL_CTX_set_session_id_context(tls_ctx->tls_ssl_ctx,
			(const unsigned char *) &server_session_id_context,
			sizeof(server_session_id_context));
		tls_ctx->tls_clt_cache = false;
		if (bitset(TLS_I_CACHE, req) && tls_cnf->tlscnf_cache_size > 0) {
			sm_log_write(tls_ctx->tls_lctx,
				TL_LCAT_INIT, TL_LMOD_CONF, SM_LOG_ERR, 6,
				"STARTTLS=%s, cache_setup=failed" , who);
			return sm_error_perm(SM_EM_TLS, EINVAL);
		}
	}

	/* load certificate locations and default CA paths */
	if (bitset(TLS_S_CERTP_OK, status) && bitset(TLS_S_CERTF_OK, status)) {
		r = SSL_CTX_load_verify_locations(tls_ctx->tls_ssl_ctx,
				tls_cnf->tlscnf_cacertfile, tls_cnf->tlscnf_cacertpath);
		if (1 == r) {
#if !TLS_NO_RSA
			if (bitset(TLS_I_RSA_TMP, req))
				SSL_CTX_set_tmp_rsa_callback(tls_ctx->tls_ssl_ctx,
					tmp_rsa_key);
#endif

			/*
			**  We have to install our own verify callback:
			**  SSL_VERIFY_PEER requests a client cert but even
			**  though *FAIL_IF* isn't set, the connection
			**  will be aborted if the client presents a cert
			**  that is not "liked" (can't be verified?) by
			**  the TLS library :-(
			*/

			/*
			**  XXX currently we could call tls_set_verify()
			**  but we hope that that function will later on
			**  only set the mode per connection.
			*/

			SSL_CTX_set_verify(tls_ctx->tls_ssl_ctx,
				bitset(TLS_I_NO_VRFY, req) ? SSL_VERIFY_NONE : SSL_VERIFY_PEER,
				tls_verify_cert_cb);

			/*
			 *  Let the verify_callback catch the verify_depth error
			 *  so that we get an appropriate error in the logfile.
			 *  SSL_CTX_set_verify(3)
			 */

			if (tls_cnf->tlscnf_vrfy_depth_isset)
				SSL_CTX_set_verify_depth(tls_ctx->tls_ssl_ctx, tls_cnf->tlscnf_vrfy_depth + 1);

			/* set up list of acceptable client certs */
			if (bitset(TLS_I_LDCLTCA, req)) {
				STACK_OF(X509_NAME) *cert_names;
				const char *calist;

				if (tls_cnf->tlscnf_calist_isset != 0) {
					if (SM_IS_EMPTY_S(tls_cnf->tlscnf_calistfile))
						calist = NULL;
					else
						calist = tls_cnf->tlscnf_calistfile;
				}
				else
						calist = tls_cnf->tlscnf_cacertfile;

				cert_names = SSL_load_client_CA_file(calist);
				if (NULL == cert_names && calist != NULL) {
					sm_log_write(tls_ctx->tls_lctx,
						TL_LCAT_INIT, TL_LMOD_CONF, SM_LOG_WARN, 9,
						"STARTTLS=%s, CAlist_file=%s, SSL_load_client_CA_file=failed",
						who, calist);
					tlslogerr(NULL, tls_ctx->tls_lctx, who);
				}
				SSL_CTX_set_client_CA_list(tls_ctx->tls_ssl_ctx, cert_names);
			}
		}
		else {
			/*
			**  can't load CA data; do we care?
			**  the data is necessary to authenticate the client,
			**  which in turn would be necessary
			**  if we want to allow relaying based on it.
			*/

			sm_log_write(tls_ctx->tls_lctx,
				TL_LCAT_INIT, TL_LMOD_CONF, SM_LOG_WARN, 7,
				"STARTTLS=%s, load_verify_locs=failed, CAcert_directory=%s, CAcert_file=%s, error=%d",
				who, tls_cnf->tlscnf_cacertpath, tls_cnf->tlscnf_cacertfile, r);
			tlslogerr(NULL, tls_ctx->tls_lctx, who);
			if (bitset(TLS_I_VRFY_LOC, req))
				return sm_error_perm(SM_EM_TLS, EINVAL);
		}
	}

/* if (!srv) SSL_CTX_set_info_callback(tls_ctx->tls_ssl_ctx, apps_ssl_info_cb); */
#if MTA_TLS_DEBUG
	if (tls_cnf->tlscnf_debug > 1)
		SSL_CTX_set_info_callback(tls_ctx->tls_ssl_ctx, apps_ssl_info_cb);
#endif

	/* install our own cipher list if set */
	param = tls_cnf->tlscnf_cipherlist;
	if (!SM_IS_EMPTY_S(param)) {
		if (SSL_CTX_set_cipher_list(tls_ctx->tls_ssl_ctx, param) <= 0) {
			sm_log_write(tls_ctx->tls_lctx,
				TL_LCAT_INIT, TL_LMOD_CONF, SM_LOG_WARN, 7,
				"STARTTLS=%s, SSL_CTX_set_cipher_list=failed, list=%s, status=ignored",
				who, param);
			tlslogerr(NULL, tls_ctx->tls_lctx, who);
			/* failure if setting to this list is required? */
		}
	}

	if (!SM_IS_EMPTY_S(tls_cnf->tlscnf_crlfile)) {
		/* get a pointer to the current certificate validation store */
		/* does not fail */
		store = SSL_CTX_get_cert_store(tls_ctx->tls_ssl_ctx);
		crl_file = BIO_new(BIO_s_file_internal());
		if (crl_file != NULL) {
			if (BIO_read_filename(crl_file, tls_cnf->tlscnf_crlfile) >= 0) {
				X509_CRL *crl;

				crl = PEM_read_bio_X509_CRL(crl_file, NULL, NULL, NULL);
				BIO_free(crl_file);
				X509_STORE_add_crl(store, crl);
				X509_CRL_free(crl);
				X509_STORE_set_flags(store, X509_V_FLAG_CRL_CHECK|X509_V_FLAG_CRL_CHECK_ALL);
				X509_STORE_set_verify_cb_func(store, x509_verify_cb);
			}
			else {
				sm_log_write(tls_ctx->tls_lctx,
					TL_LCAT_INIT, TL_LMOD_CONF, SM_LOG_WARN, 7,
					"STARTTLS=%s, CRLfile=%s, PEM_read_bio_X509_CRL=failed",
					who, tls_cnf->tlscnf_crlfile);

				/* avoid memory leaks */
				BIO_free(crl_file);
				ok = false;
			}

		}
		else {
			sm_log_write(tls_ctx->tls_lctx,
				TL_LCAT_INIT, TL_LMOD_CONF, SM_LOG_WARN, 7,
				"STARTTLS=%s, CRLfile=%s, BIO_new=failed",
				who, tls_cnf->tlscnf_crlfile);
			ok = false;
		}
	}
	else
		store = NULL;
	if (!SM_IS_EMPTY_S(tls_cnf->tlscnf_crlpath) && store != NULL) {
		X509_LOOKUP *x509_lookup;

		x509_lookup = X509_STORE_add_lookup(store, X509_LOOKUP_hash_dir());
		if (NULL == x509_lookup) {
			sm_log_write(tls_ctx->tls_lctx,
				TL_LCAT_INIT, TL_LMOD_CONF, SM_LOG_WARN, 7,
				"STARTTLS=%s, CRLpath=%s, X509_STORE_add_lookup(hash)=failed",
				who, tls_cnf->tlscnf_crlpath);
			ok = false;
		}
		else {
			X509_LOOKUP_add_dir(x509_lookup, tls_cnf->tlscnf_crlpath, X509_FILETYPE_PEM);
			X509_STORE_set_flags(store, X509_V_FLAG_CRL_CHECK|X509_V_FLAG_CRL_CHECK_ALL);
		}
	}

	return ok ? SM_SUCCESS : sm_error_perm(SM_EM_TLS, EINVAL);
}

/*
**  TLS_GET_INFO -- get information about TLS connection
**
**	Parameters:
**		tls_ctx -- TLS library context
**		ssl -- TLS connection structure
**		flags -- server or client, and other flags
**		tlsi_ctx -- TLS information context
**
**	Returns:
**		usual error code
**
**	Side Effects:
**		sets data in tlsi_ctx
*/

sm_ret_T
tls_get_info(tls_ctx_P tls_ctx, SSL *ssl, uint flags, tlsi_ctx_P tlsi_ctx)
{
	int b, r, i;
	char *s;
	const SSL_CIPHER *c;
	X509 *cert;

	SM_REQUIRE(ssl != NULL);
	SM_REQUIRE(tlsi_ctx != NULL);
	c = SSL_get_current_cipher(ssl);

#define TLS_CERT_RQST	(SM_IS_FLAG(flags, TLS_F_CERT_RQST))

	s = (char *) SSL_CIPHER_get_name(c);
	TLS_SET_STR(s, tlsi_ctx->tlsi_cipher, 32);

	b = SSL_CIPHER_get_bits(c, &r);
	tlsi_ctx->tlsi_cipher_bits = b;
	tlsi_ctx->tlsi_algs_bits = r;
	tlsi_ctx->tlsi_reused = SSL_session_reused(ssl);

	s = SSL_CIPHER_get_version(c);
	TLS_SET_STR(s, tlsi_ctx->tlsi_version, 16);

	cert = SSL_get_peer_certificate(ssl);
	if (cert != NULL) {
		uint n;
		uchar md[EVP_MAX_MD_SIZE];
		X509_NAME *subj, *issuer;
		char buf[SM_X509_NAME_MAX];

		subj = X509_get_subject_name(cert);
		issuer = X509_get_issuer_name(cert);

		X509_NAME_oneline(subj, buf, sizeof buf);
		TLS_SET_XSTR(buf, tlsi_ctx->tlsi_cert_subject);
		X509_NAME_oneline(issuer, buf, sizeof buf);
		TLS_SET_XSTR(buf, tlsi_ctx->tlsi_cert_issuer);

		(void) sm_tls_get_cn(subj, NID_commonName, &tlsi_ctx->tlsi_cn_subject);
		(void) sm_tls_get_cn(issuer, NID_commonName, &tlsi_ctx->tlsi_cn_issuer);
		r = sm_tls_get_altname(cert, tlsi_ctx->tlsi_altnames, SM_ARRAY_SIZE(tlsi_ctx->tlsi_altnames));
		if (r >= 0)
			tlsi_ctx->tlsi_naltnames = r;
		else
			tlsi_ctx->tlsi_naltnames = -1;
		n = 0;

		if (X509_digest(cert, EVP_DIGEST, md, &n) != 0 && n > 0) {
			sm_str_P str;

#if FFR_CERT_PINNING
			sm_memcpy(tlsi_ctx->tlsi_cert_md, md, sizeof(tlsi_ctx->tlsi_cert_md));
			tlsi_ctx->tlsi_cert_mdlen = n;
#endif
			r = (n * 3) + 2;
			if (NULL == tlsi_ctx->tlsi_cert_fp && r > n)
				tlsi_ctx->tlsi_cert_fp = sm_str_new(NULL, r, r + 2);
			str = tlsi_ctx->tlsi_cert_fp;
			if (str != NULL) {
				sm_str_clr(str);
				sm_strprintf(str, "%0.*hhy", n, md);
			}
		}
		else
			TLS_SET_STR("", tlsi_ctx->tlsi_cert_fp, 256);

#if MTA_TLS_NOTAFTER
#define DATELEN sizeof("YYYYmmddHHMMSSZ")
		tlsi_ctx->tlsi_notAfter = sm_str_new(NULL, DATELEN, DATELEN + 2);
		if (tlsi_ctx->tlsi_notAfter != NULL) {
			ASN1_UTCTIME *tm;

			tm = X509_get_notAfter(cert);
    		s = tm->data;
			sm_str_clr(tlsi_ctx->tlsi_notAfter);
			if (strlen(s) < DATELEN)	/* HACK!!!!! */
				(void) sm_str_scat(tlsi_ctx->tlsi_notAfter, "20");
			(void) sm_str_scat(tlsi_ctx->tlsi_notAfter, s);
		}
#endif /* MTA_TLS_NOTAFTER */
#if FFR_CERT_PINNING || MTA_TLS_NOTAFTER
		tlsi_ctx->tlsi_notAfter_T = ASN1_GetTimeT(X509_get_notAfter(cert));
#endif
	}
	else {
		TLS_SET_STR("", tlsi_ctx->tlsi_cert_subject, SM_TLS_NAME_MAX);
		TLS_SET_STR("", tlsi_ctx->tlsi_cert_issuer, SM_TLS_NAME_MAX);
		TLS_SET_STR("", tlsi_ctx->tlsi_cn_subject, SM_TLS_NAME_MAX);
		TLS_SET_STR("", tlsi_ctx->tlsi_cn_issuer, SM_TLS_NAME_MAX);
		for (i = 0; i < SM_ARRAY_SIZE(tlsi_ctx->tlsi_altnames); i++)
			TLS_SET_STR("", tlsi_ctx->tlsi_altnames[i], SM_TLS_NAME_MAX);
		TLS_SET_STR("", tlsi_ctx->tlsi_cert_fp, 256);
	}

	switch (SSL_get_verify_result(ssl)) {
	  case X509_V_OK:
		if (cert != NULL)
			r = TLS_VRFY_OK;
		else if (TLS_CERT_RQST)
			r = TLS_VRFY_NO;
		else
			r = TLS_VRFY_NOTR;
		break;
	  default:
		r = TLS_VRFY_FAIL;
		break;
	}
	tlsi_ctx->tlsi_vrfy = r;
	if (cert != NULL)
		X509_free(cert);

	return SM_SUCCESS;
}

#if !TLS_NO_RSA
/*
**  TMP_RSA_KEY -- return temporary RSA key
**
**	Parameters:
**		s -- TLS connection structure
**		export --
**		keylength --
**
**	Returns:
**		temporary RSA key.
*/

# ifndef RSA_TMP_CNT_MAX
#  define RSA_TMP_CNT_MAX	100	/* XXX better value? */
# endif

/* ARGUSED0 */
static RSA *
tmp_rsa_key(SSL *ssl, int export, int keylength)
{
	static int RSATmpCnt = 0;

	/*
	**  NOT thread safe!
	**  for thread safety: see comment at begin of file.
	**  Note: currently this is not used by any program that uses
	**  POSIX threads (only statethreads).
	*/

	if (++RSATmpCnt < RSA_TMP_CNT_MAX)
		return rsa_tmp;
	RSATmpCnt = 0;
	if (rsa_tmp != NULL)
		RSA_free(rsa_tmp);
	rsa_tmp = RSA_generate_key(RSA_KEYLENGTH, RSA_F4, NULL, NULL);
	if (NULL == rsa_tmp) {
		/* COMPLAIN, but HOW? */
	}
	return rsa_tmp;
}
#endif /* !TLS_NO_RSA */

/*
**  TLS_VRFY2TXT -- return textual representation of verification result
**
**	Parameters:
**		vrfy -- verification result
**
**	Returns:
**		textual representation of verification result
*/

const char *
tls_vrfy2txt(int vrfy)
{
	switch (vrfy) {
	  case TLS_VRFY_OK:
		return "OK";
	  case TLS_VRFY_NO:
		return "NO";
	  case TLS_VRFY_NOTR:
		return "NOT";
	  case TLS_VRFY_FAIL:
		return "FAIL";
	  default:
		return "Unknown";
	}
	/* NOTREACHED */
	return "OOPS";
}

/*
**  TLS_SET_MD -- set MD based on conf data
**
**	Parameters:
**		tls_ctx -- TLS library context
**		tls_cnf -- TLS configuration data
*/

sm_ret_T
tls_set_md(tls_ctx_P tls_ctx, tls_cnf_P tls_cnf)
{
	char *fp_dig;
	const EVP_MD *md;

	SM_REQUIRE(tls_cnf != NULL);

	if (0 == tls_cnf->tlscnf_fp_dig_isset)
		return SM_NOTDONE;
	fp_dig = tls_cnf->tlscnf_fp_dig;
	if (SM_IS_EMPTY_S(fp_dig))
		return SM_NOTDONE;

	md = EVP_get_digestbyname(fp_dig);
	if (NULL == md)
		return sm_error_perm(SM_EM_TLS, EINVAL);
	
	tls_ctx->tls_EVP_digest = md;
	return SM_SUCCESS;
}
#endif /* MTA_USE_TLS */
