/*
 * Copyright (c) 2005 Sendmail, Inc. and its suppliers.
 *	All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: socklisten.c,v 1.6 2011/06/03 23:19:07 ca Exp $")
#include "sm/assert.h"
#include "sm/error.h"
#include "sm/memops.h"
#include "sm/pwd.h"
#include "sm/grp.h"
#include "sm/fcntl.h"
#include "sm/net.h"
#include "sm/sockcnf.h"
#include "sm/socket.h"

/*
**  SOCK_LISTEN -- Create a socket, bind, and listen.
**
**	Parameters:
**		sock -- address to bind to.
**		backlog -- Backlog of connections to accept (man listen(2))
**		flags -- flags
**		pfd -- (pointer to) file descriptor (output)
**
**	Returns:
**		usual error code
*/

sm_ret_T
sock_listen(sockspec_P sock, int backlog, uint flags, int *pfd)
{
	sm_ret_T ret;
	int listenfd, sockopt, r, save_errno;
	mode_t mode;
	sockaddr_len_T sock_size;
	sm_sockaddr_T sm_sockaddr;

	SM_ASSERT(sock != NULL);
	SM_ASSERT(pfd != NULL);
	SM_ASSERT(backlog > 0);

	mode = (mode_t) -1;
	listenfd = INVALID_SOCKET;
	ret = sockspec2sock(sock, &sm_sockaddr, &sock_size, flags);
	if (sm_is_err(ret))
		goto error;

	sockopt = 1;
	listenfd = socket(sm_sockaddr.sa.sa_family, SOCK_STREAM, 0);
	if (listenfd == INVALID_SOCKET)
		return sm_error_perm(SM_EM_NET, errno);
	if (setsockopt(listenfd, SOL_SOCKET, SO_REUSEADDR, (void *) &sockopt,
		       sizeof(sockopt)) == -1)
	{
		ret = sm_error_perm(SM_EM_NET, errno);
		goto error;
	}

	if (sm_sockaddr.sa.sa_family == AF_UNIX)
		mode = umask(sock->sock_unix.unixsckspc_umask);
	r = bind(listenfd, (struct sockaddr *) &sm_sockaddr, sock_size);
	save_errno = errno;
	if (mode != (mode_t) -1 && sm_sockaddr.sa.sa_family == AF_UNIX)
		(void) umask(mode);
	if (r == -1) {
		ret = sm_error_perm(SM_EM_NET, save_errno);
		goto error;
	}
	if (sm_sockaddr.sa.sa_family == AF_UNIX) {
		struct passwd  *pwd;
		struct group   *grp;

		/*
		**  Some OS (e.g., HP UX) don't obey umask for bind().
		**  Of course this has a race condition, hence the socket
		**  MUST be in a "safe" directory (which currently is NOT
		**  checked).
		*/

		mode = 0777 ^ sock->sock_unix.unixsckspc_umask;
		r = chmod(sm_sockaddr.sunix.sun_path, mode);
		save_errno = errno;
		if (r != 0) {
			ret = sm_error_perm(SM_EM_NET, save_errno);
			goto error;
		}

		pwd = NULL;
		if (sock->sock_unix.unixsckspc_user != NULL &&
		    *sock->sock_unix.unixsckspc_user != '\0' &&
		    (pwd = getpwnam(sock->sock_unix.unixsckspc_user)) == NULL)
		{
			ret = sm_error_perm(SM_EM_NET, EINVAL);
			goto error;
		}
		if (pwd != NULL) {
			grp = NULL;
			if (sock->sock_unix.unixsckspc_group != NULL &&
			    *sock->sock_unix.unixsckspc_group != '\0' &&
			    (grp = getgrnam(sock->sock_unix.unixsckspc_group)) == NULL)
			{
				ret = sm_error_perm(SM_EM_NET, EINVAL);
				goto error;
			}
			if (grp != NULL)
				pwd->pw_gid = grp->gr_gid;

			r = chown(sm_sockaddr.sunix.sun_path, pwd->pw_uid, pwd->pw_gid);
			save_errno = errno;
			if (r != 0) {
				ret = sm_error_perm(SM_EM_NET, save_errno);
				goto error;
			}
		}
	}

	if (listen(listenfd, backlog) == -1) {
		ret = sm_error_perm(SM_EM_NET, errno);
		goto error;
	}
	*pfd = listenfd;
	return SM_SUCCESS;

  error:
	if (listenfd != INVALID_SOCKET)
		closesocket(listenfd);
	return ret;
}
