/*
 *  simple producer/consumer program using pthreads.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: t-prdcns-0.c,v 1.2 2010/09/08 02:18:12 ca Exp $")

#include "sm/assert.h"
#include "sm/error.h"
#include "sm/memops.h"
#include "sm/heap.h"
/* #include "sm/test.h" */
#include "sm/fxszq.h"

#include <stdio.h>
#include <pthread.h>

/*
*/

struct fxszq_S
{
	uint	 fsq_max;   /* allocated size */
	uint	 fsq_cur;   /* currently used (basically last-first) */
	uint	 fsq_first; /* first index to read */
	uint	 fsq_last;  /* last index to read (first to write) */
	int	*fsq_array; /* array */
};
typedef struct fxszq_S	fxszq_T, *fxszq_P;
fxszq_T q1;

#define WHAT_TERM	0
#define WHAT_CONT	1

#define REPS 5

#define T_QS	256

static int Verbose = 0;
static long reps = 0;
static pthread_t task1, task2;
static bool stop = false;

struct t_ctx_S
{
	int			 t_called;
	int			 t_once;
	pthread_t 	*t_wakeup;
};
typedef struct t_ctx_S t_ctx_T, *t_ctx_P;

pthread_mutex_t	 q_mutex;
pthread_cond_t	 q_cond;
long q_add, q_rm;

#define SM_TEST(x)	assert(x)
#define SM_TEST_E(x)	assert(x)

static void
t_init(uint sz)
{
	int r;

	SM_TEST(sz != 0);
	if (sz == 0)
		return;
	q1.fsq_array = sm_malloc(sizeof(int) * sz);
	SM_TEST(q1.fsq_array != NULL);
	if (q1.fsq_array == NULL)
		return;
	FSQ_INIT(sz, q1.fsq_max, q1.fsq_cur, q1.fsq_first, q1.fsq_last);
	SM_TEST(FSQ_IS_EMPTY(q1.fsq_cur));
	SM_TEST(!FSQ_IS_FULL(q1.fsq_max, q1.fsq_cur));
	r = pthread_mutex_init(&q_mutex, NULL);
	SM_TEST(r == 0);
	r = pthread_cond_init(&q_cond, NULL);
	SM_TEST(r == 0);
}

static bool
add_q(int i)
{
	int r;
	bool e;

	r = pthread_mutex_lock(&q_mutex);
	if (FSQ_IS_FULL(q1.fsq_max, q1.fsq_cur)) {
		r = pthread_mutex_unlock(&q_mutex);
		return true;
	}
	FSQ_APPEND(q1.fsq_max, q1.fsq_last, q1.fsq_cur, q1.fsq_array, i);
	SM_TEST(!FSQ_IS_EMPTY(q1.fsq_cur));
	++q_add;
	if (q_add > reps) stop = true;
	e = FSQ_IS_FULL(q1.fsq_max, q1.fsq_cur);
	r = pthread_mutex_unlock(&q_mutex);
	return e;
}

static bool
rm_q(void)
{
	int j, h, r;
	bool e;

	r = pthread_mutex_lock(&q_mutex);
	if (FSQ_IS_EMPTY(q1.fsq_cur)) {
		r = pthread_mutex_unlock(&q_mutex);
		return true;
	}
	j = FSQ_GET(q1.fsq_max, q1.fsq_first, q1.fsq_cur, q1.fsq_array, h);
	++q_rm;
	e = FSQ_IS_EMPTY(q1.fsq_cur);
	r = pthread_mutex_unlock(&q_mutex);
	return e;
}

static void *
t_add(void *ctx)
{
	int i;
	bool e;
	t_ctx_P fctx;
	struct timespec w;
	struct timeval now;

	SM_ASSERT(ctx != NULL);
	fctx = (t_ctx_P) ctx;

	while (1) {
		if (Verbose > 0)
			fprintf(stderr, "t_add=%d, reps=%ld\n", fctx->t_called, reps);
		fctx->t_called++;
		if (stop) {
			i = pthread_cond_signal(&q_cond);
			return 0;
		}

		for (i = 0; i < fctx->t_once && !(e = add_q(i)); i++)
			;
		if (Verbose > 0)
			fprintf(stderr, "t_add: e=%d, stop=%d\n", e, stop);
		if (!stop) {
			i = pthread_cond_signal(&q_cond);
			SM_TEST_E(i == 0);
		}
		if (!e)
			continue;

		if (Verbose > 0)
			fprintf(stderr, "t_add: beforelock=%d\n", i);
		i = pthread_mutex_lock(&q_mutex);
		SM_TEST_E(i == 0);
		if (Verbose > 0)
			fprintf(stderr, "t_add: afterlock=%d\n", i);
		while (FSQ_IS_FULL(q1.fsq_max, q1.fsq_cur) && !stop) {
			gettimeofday(&now, NULL);
			w.tv_sec = now.tv_sec + 2;
			w.tv_nsec = 0;
			i = pthread_cond_timedwait(&q_cond, &q_mutex, &w);
			if (stop) return 0;
			if (i != 0) {
				fprintf(stderr, "t_add: afterwait=%d, full=%d, called=%d\n",
					i, FSQ_IS_FULL(q1.fsq_max, q1.fsq_cur), fctx->t_called);
			}
			SM_TEST_E(i == 0);
			if (Verbose > 0)
				fprintf(stderr, "t_add: afterwait=%d, full=%d\n", i,
					FSQ_IS_FULL(q1.fsq_max, q1.fsq_cur));
		}
		i = pthread_mutex_unlock(&q_mutex);
		SM_TEST_E(i == 0);
		if (stop) return 0;
	}
}

static void *
t_rm(void *ctx)
{
	int i;
	bool e;
	t_ctx_P fctx;
	struct timespec w;
	struct timeval now;

	SM_ASSERT(ctx != NULL);
	fctx = (t_ctx_P) ctx;

	while (1) {
		if (Verbose > 0)
			fprintf(stderr, "	t_rm=%d, reps=%ld\n", fctx->t_called, reps);
		fctx->t_called++;
		if (stop) {
			i = pthread_cond_signal(&q_cond);
			return 0;
		}

		for (i = 0; i < fctx->t_once && !(e = rm_q()); i++)
			;
		if (Verbose > 0)
			fprintf(stderr, "	t_rm: e=%d, stop=%d\n", e, stop);
		if (!stop) {
			i = pthread_cond_signal(&q_cond);
			SM_TEST_E(i == 0);
		}
		if (!e)
			continue;

		if (Verbose > 0)
			fprintf(stderr, "	t_rm: beforelock=%d\n", i);
		i = pthread_mutex_lock(&q_mutex);
		SM_TEST_E(i == 0);
		if (Verbose > 0)
			fprintf(stderr, "	t_rm: afterlock=%d\n", i);
		while (FSQ_IS_EMPTY(q1.fsq_cur) && !stop) {
			gettimeofday(&now, NULL);
			w.tv_sec = now.tv_sec + 2;
			w.tv_nsec = 0;
			i = pthread_cond_timedwait(&q_cond, &q_mutex, &w);
			if (stop) return 0;
			if (i != 0) {
				fprintf(stderr, "t_rm: afterwait=%d, empty=%d, called=%d\n",
					i, FSQ_IS_EMPTY(q1.fsq_cur), fctx->t_called);
			}
			SM_TEST_E(i == 0);
			if (Verbose > 0)
				fprintf(stderr, "	t_rm: afterwait=%d, empty=%d\n", i,
					FSQ_IS_EMPTY(q1.fsq_cur));
		}
		i = pthread_mutex_unlock(&q_mutex);
		SM_TEST_E(i == 0);
		if (stop) return 0;
	}
}

static void
testev(uint sz, int adds, int rms)
{
	int ret;
	t_ctx_T tctx1, tctx2;

	t_init(sz);


	tctx1.t_called = 0;
	tctx1.t_once = adds;
	ret = pthread_create(&task1, NULL, t_add, (void *) &tctx1);
	SM_TEST(ret == 0);
	SM_TEST(task1 != NULL);

	tctx2.t_called = 0;
	tctx2.t_once = rms;
	ret = pthread_create(&task2, NULL, t_rm, (void *) &tctx2);
	SM_TEST(ret == 0);
	SM_TEST(task2 != NULL);

	tctx1.t_wakeup = &task2;
	tctx2.t_wakeup = &task1;

	ret = pthread_join(task1, NULL);
	/* SM_TEST(ret == 0); */
	fprintf(stderr, "join1=%d\n", ret);
	ret = pthread_join(task2, NULL);
	/* SM_TEST(ret == 0); */
	fprintf(stderr, "join2=%d\n", ret);

	fprintf(stderr, "adds=%ld, rms=%ld\n", q_add, q_rm);
}

static void
usage(const char *prg)
{
	fprintf(stderr, "%s: simple producer/consumer test\n", prg);
	fprintf(stderr, "usage %s [options]\n", prg);
	fprintf(stderr, "options\n");
	fprintf(stderr, "-a n  add n items on each iteration\n");
	fprintf(stderr, "-d n  remove n items on each iteration\n");
	fprintf(stderr, "-r n  add n items in total\n");
	fprintf(stderr, "-s n  set size of queue to n\n");
	fprintf(stderr, "-V    increase verbosity\n");
}

int
main(int argc, char *argv[])
{
	int c, adds, rms;
	uint sz;

	reps = 1;
	sz = 100;
	adds = rms = 10;
	while ((c = getopt(argc, argv, "a:d:r:s:V")) != -1)
	{
		switch (c)
		{
		  case 'a':
			adds = atoi(optarg);
			break;
		  case 'd':
			rms = atoi(optarg);
			break;
		  case 'r':
			reps = atoi(optarg);
			break;
		  case 's':
			sz = (uint) atoi(optarg);
			break;
		  case 'V':
			Verbose++;
			break;
		  default:
			usage(argv[0]);
			return(1);
		}
	}
/*
	sm_test_begin(argc, argv, "test evthr perf 0");
*/
	testev(sz, adds, rms);
	return 0;
/*
	return sm_test_end();
*/
}
